import pytest
from unittest import mock
from datetime import datetime
try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from django_forest.resources.utils.queryset.filters.date.conditions import (
    RangeCondition,
    LowerThanCondition,
    LowerThanOrEqualCondition,
    GreaterThanCondition,
    GreaterThanOrEqualCondition,
)
from django_forest.resources.utils.queryset.filters.date.factory import ConditionFactory


def _init_factory(dt: datetime):
    return ConditionFactory(dt)

def test_date_condition_builders():
    assert ConditionFactory.DATE_CONDITION_BUILDERS == {
        'today': 'build_today_condition',
        'yesterday': 'build_yesterday_condition',
        'previous_x_days': 'build_previous_x_days_condition',
        'previous_x_days_to_date': 'build_previous_x_days_to_date_condition',
        'previous_week': 'build_previous_week_condition',
        'previous_week_to_date': 'build_previous_week_to_date_condition',
        'previous_month': 'build_previous_month_condition',
        'previous_month_to_date': 'build_previous_month_to_date_condition',
        'previous_quarter': 'build_previous_quarter_condition',
        'previous_quarter_to_date': 'build_previous_quarter_to_date_condition',
        'previous_year': 'build_previous_year_condition',
        'previous_year_to_date': 'build_previous_year_to_date_condition',
        'before_x_hours_ago': 'build_before_x_hours_ago_condition',
        'after_x_hours_ago': 'build_after_x_hours_ago_condition',
        'past': 'build_past_condition',
        'future': 'build_future_condition',
    }

def test_operators():
    assert ConditionFactory.OPERATORS == [
        'today',
        'yesterday',
        'previous_x_days',
        'previous_x_days_to_date',
        'previous_week',
        'previous_week_to_date',
        'previous_month',
        'previous_month_to_date',
        'previous_quarter',
        'previous_quarter_to_date',
        'previous_year',
        'previous_year_to_date',
        'before_x_hours_ago',
        'after_x_hours_ago',
        'past',
        'future',
    ]

def test_offset_operators():
    assert ConditionFactory.OFFSET_OPERATORS == [
        'today',
        'yesterday',
        'previous_x_days',
        'previous_x_days_to_date',
        'previous_week',
        'previous_week_to_date',
        'previous_month',
        'previous_month_to_date',
        'previous_quarter',
        'previous_quarter_to_date',
        'previous_year',
        'previous_year_to_date',
    ]

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2022, 1, 1),
                datetime(2022, 1, 2)
            ),
            0
        ),
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2021, 12, 31),
                datetime(2022, 1, 1)
            ),
            1
        ),
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2021, 12, 30),
                datetime(2021, 12, 31)
            ),
            2
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2023, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC"))
            ),
            0
        ),
        (
            datetime(2020, 2, 29, tzinfo=zoneinfo.ZoneInfo("America/Araguaina")), 
            RangeCondition(
                datetime(2020, 2, 29, tzinfo=zoneinfo.ZoneInfo("America/Araguaina")),
                datetime(2020, 3, 1, tzinfo=zoneinfo.ZoneInfo("America/Araguaina"))
            ),
            0
        ),
        (
            datetime(2025, 8, 10, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2025, 8, 10, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2025, 8, 11, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
            0
        ),
    )
)
def test_build_today_condition(current: datetime, expected: RangeCondition, offset: int) -> None:
    factory = _init_factory(current)
    assert factory.build_today_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                start=datetime(2021, 12, 31),
                end=datetime(2022, 1, 1)
            ),
            0
        ),
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                start=datetime(2021, 12, 30),
                end=datetime(2021, 12, 31)
            ),
            1
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                start=datetime(2022, 12, 30, tzinfo=zoneinfo.ZoneInfo("UTC")),
                end=datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC"))
            ),
            0
        ),
        (
            datetime(2020, 2, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                start=datetime(2020, 1, 31, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                end=datetime(2020, 2, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
            0
        ),
        (
            datetime(2020, 2, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                start=datetime(2020, 1, 30, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                end=datetime(2020, 1, 31, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
            1
        ),
        (
            datetime(2025, 8, 10, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                start=datetime(2025, 8, 9, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                end=datetime(2025, 8, 10, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
            0
        ),
    )
)
def test_build_yesterday_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_yesterday_condition(current, offset) == expected


@pytest.mark.parametrize(
    "current,days,expected,offset",
    (   
        (
            datetime(2022, 1, 1, 23, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            1,
            RangeCondition(
                datetime(2021, 12, 31, 0, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 1, 0, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe"))
            ),
            0
        ),
        (
            datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            1,
            RangeCondition(
                datetime(2021, 12, 30, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2021, 12, 31, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe"))
            ),
            1
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            2,
            RangeCondition(
                datetime(2022, 12, 29, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo"))
            ),
            0
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            2,
            RangeCondition(
                datetime(2022, 12, 27, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 12, 29, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo"))
            ),
            1
        ),
        (
            datetime(2020, 2, 1, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            5,
            RangeCondition(
                datetime(2020, 1, 27, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2020, 2, 1, tzinfo=zoneinfo.ZoneInfo("UTC"))
            ),
            0
        ),
        (
            datetime(2025, 8, 10), 
            60,
            RangeCondition(
                datetime(2025, 6, 11),
                datetime(2025, 8, 10)
            ),
            0
        ),
    )
)
def test_build_previous_x_days_condition(current, days, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_x_days_condition(current, offset, days) == expected

@pytest.mark.parametrize(
    "current,days,expected,offset",
    (   
        (
            datetime(2022, 1, 1), 
            1,
            RangeCondition(
                datetime(2022, 1, 1),
                datetime(2022, 1, 2)
            ),
            0
        ),
        (
            datetime(2022, 1, 1), 
            1,
            RangeCondition(
                datetime(2021, 12, 31),
                datetime(2022, 1, 1)
            ),
            1
        ),
        (
            datetime(2022, 1, 1), 
            1,
            RangeCondition(
                datetime(2021, 12, 30),
                datetime(2021, 12, 31)
            ),
            2
        ),
        (
            datetime(2022, 12, 31), 
            2,
            RangeCondition(
                datetime(2022, 12, 30),
                datetime(2023, 1, 1)
            ),
            0
        ),
        (
            datetime(2020, 2, 1), 
            5,
            RangeCondition(
                datetime(2020, 1, 28),
                datetime(2020, 2, 2)
            ),
            0
        ),
        (
            datetime(2020, 2, 1), 
            5,
            RangeCondition(
                datetime(2020, 1, 23),
                datetime(2020, 1, 28)
            ),
            1
        ),
        (
            datetime(2020, 2, 1), 
            5,
            RangeCondition(
                datetime(2020, 1, 23),
                datetime(2020, 1, 28)
            ),
            1
        ),
        (
            datetime(2025, 8, 10), 
            60,
            RangeCondition(
                datetime(2025, 6, 12),
                datetime(2025, 8, 11)
            ),
            0
        ),
    )
)
def test_build_previous_x_days_to_date_condition(current, days, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_x_days_to_date_condition(current, offset, days) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (
        (
            datetime(2022, 1, 5), 
            RangeCondition(
                datetime(2021, 12, 27),
                datetime(2022, 1, 3)
            ),
            0
        ),
        (
            datetime(2022, 1, 5), 
            RangeCondition(
                datetime(2021, 12, 20),
                datetime(2021, 12, 27)
            ),
            1
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2022, 2, 7, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe"))
            ),
            0
        ),
        (
            datetime(2022, 4, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 4, 4, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 4, 11, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2017, 11, 6, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2017, 10, 30, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2017, 11, 6, tzinfo=zoneinfo.ZoneInfo("UTC"))
            ),
            0
        ),
        (
            datetime(2017, 11, 6, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2017, 10, 16, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2017, 10, 23, tzinfo=zoneinfo.ZoneInfo("UTC"))
            ),
            2
        ),
    )
)
def test_build_previous_week_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_week_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2022, 1, 3, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 6, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            ),
            0
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 2, 15, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 2, 7, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            1
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 1, 31, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 2, 7, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            2
        ),
        (
            datetime(2022, 4, 14, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 4, 11, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 4, 15, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2017, 11, 6), 
            RangeCondition(
                datetime(2017, 11, 6),
                datetime(2017, 11, 7),
            ),
            0
        ),
    )
)
def test_previous_week_to_date_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_week_to_date_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2021, 12, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            ),
            0
        ),
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2021, 11, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2021, 12, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            ),
            1
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 2, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 4, 14, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 3, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 4, 14, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 2, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            2
        ),
        (
            datetime(2017, 11, 6), 
            RangeCondition(
                datetime(2017, 10, 1),
                datetime(2017, 11, 1),
            ),
            0
        ),
    )
)
def test_build_previous_month_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_month_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 1, 6, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            ),
            0
        ),
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2021, 12, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            ),
            1
        ),
        (
            datetime(2022, 1, 5, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2021, 11, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2021, 12, 1, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            ),
            2
        ),
        (
            datetime(2022, 2, 14, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 2, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 2, 15, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 4, 14, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 4, 15, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
            ),
            0
        ),
        (
            datetime(2017, 11, 1), 
            RangeCondition(
                datetime(2017, 11, 1),
                datetime(2017, 11, 2),
            ),
            0
        ),
    )
)
def test_build_previous_month_to_date_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_month_to_date_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2021, 10, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
            ),
            0
        ),
        (
            datetime(2022, 2, 12, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2021, 10, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
            ),
            0
        ),
        (
            datetime(2022, 3, 31, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")), 
            RangeCondition(
                datetime(2021, 10, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Aqtobe")),
            ),
            0
        ),
        (
            datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 6, 30, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Antarctica/McMurdo")),
            ),
            0
        ),
        (
            datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 8, 12, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 9, 30, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 9, 30, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            1
        ),
        (
            datetime(2022, 10, 1), 
            RangeCondition(
                datetime(2022, 7, 1),
                datetime(2022, 10, 1),
            ),
            0
        ),
        (
            datetime(2022, 11, 12), 
            RangeCondition(
                datetime(2022, 7, 1),
                datetime(2022, 10, 1),
            ),
            0
        ),
        (
            datetime(2022, 12, 31), 
            RangeCondition(
                datetime(2022, 7, 1),
                datetime(2022, 10, 1),
            ),
            0
        ),
        (
            datetime(2022, 12, 31), 
            RangeCondition(
                datetime(2022, 1, 1),
                datetime(2022, 4, 1),
            ),
            2
        ),
    )
)
def test_build_previous_quarter_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_quarter_condition(current, offset) == expected 

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
                datetime(2022, 1, 2, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            ),
            0
        ),
        (
            datetime(2022, 2, 12, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
                datetime(2022, 2, 13, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")), 
            ),
            0
        ),
        (
            datetime(2022, 3, 31, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            ),
            0
        ),
        (
            datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2022, 4, 2, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
            0
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2022, 5, 13, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
            0
        ),
        (
            datetime(2022, 6, 30, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2022, 4, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
            0
        ),
        (
            datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 7, 2, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 8, 12, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 8, 13, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 9, 30, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 7, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 10, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 10, 1), 
            RangeCondition(
                datetime(2022, 10, 1),
                datetime(2022, 10, 2),
            ),
            0
        ),
        (
            datetime(2022, 11, 12), 
            RangeCondition(
                datetime(2022, 10, 1),
                datetime(2022, 11, 13),
            ),
            0
        ),
        (
            datetime(2022, 12, 31), 
            RangeCondition(
                datetime(2022, 10, 1),
                datetime(2023, 1, 1),
            ),
            0
        ),
    )
)
def test_build_previous_quarter_to_date_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_quarter_to_date_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2021, 1, 1),
                datetime(2022, 1, 1),
            ),
            0
        ),
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2020, 1, 1),
                datetime(2021, 1, 1),
            ),
            1
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2021, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2019, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2020, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            2
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2021, 1, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
            0
        ),
    )
)
def test_previous_year_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_year_condition(current, offset) == expected

@pytest.mark.parametrize(
    "current,expected,offset",
    (   
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2022, 1, 1),
                datetime(2022, 1, 2),
            ),
            0
        ),
        (
            datetime(2022, 1, 1), 
            RangeCondition(
                datetime(2021, 1, 1),
                datetime(2022, 1, 1),
            ),
            1
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
                datetime(2022, 5, 13, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
            0
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2022, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2023, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            0
        ),
        (
            datetime(2022, 12, 31, tzinfo=zoneinfo.ZoneInfo("UTC")), 
            RangeCondition(
                datetime(2020, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
                datetime(2021, 1, 1, tzinfo=zoneinfo.ZoneInfo("UTC")),
            ),
            2
        ),
    )
)
def test_build_previous_year_to_date_condition(current, expected, offset):
    factory = _init_factory(current)
    assert factory.build_previous_year_to_date_condition(current, offset) == expected


@pytest.mark.parametrize(
    "current,value,expected",
    (   
        (
            datetime(2022, 1, 1),
            2,
            LowerThanCondition(
                datetime(2021, 12, 31, 22),
            ),
        ),
        (
            datetime(2022, 5, 12, 0, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            1,
            LowerThanCondition(
                datetime(2022, 5, 11, 23, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
        ),
        (
            datetime(2022, 5, 12, 0, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            1,
            LowerThanCondition(
                datetime(2022, 5, 11, 23, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
        ),
        (
            datetime(2022, 2, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            1,
            LowerThanCondition(
                datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            ),
        ),
    )
)
def test_build_before_x_hours_ago_condition(current, value, expected):
    factory = _init_factory(current)
    assert factory.build_before_x_hours_ago_condition(current, value) == expected

@pytest.mark.parametrize(
    "current,value,expected",
    (   
        (
            datetime(2021, 12, 31, 23),
            2,
            GreaterThanCondition(
                datetime(2022, 1, 1, 1)
            ),
        ),
        (
            datetime(2022, 5, 12, 0, 9, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            1,
            GreaterThanCondition(
                datetime(2022, 5, 12, 1, 9, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
        ),
        (
            datetime(2022, 5, 12, 0, 9, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            2,
            GreaterThanCondition(
                datetime(2022, 5, 12, 2, 9, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            ),
        ),
        (
            datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            1,
            GreaterThanCondition(
                datetime(2022, 2, 1, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo"))
            ),
        ),
    )
)
def test_build_after_x_hours_ago_condition(current, value, expected):
    factory = _init_factory(current)
    assert factory.build_after_x_hours_ago_condition(current, value) == expected

@pytest.mark.parametrize(
    "current,expected",
    (   
        (
            datetime(2021, 12, 31, 23),
            LowerThanOrEqualCondition(datetime(2021, 12, 31, 23))
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            LowerThanOrEqualCondition(datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")))
        ),
        (
            datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            LowerThanOrEqualCondition(datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")))
        ),
    )
)
def test_build_past_condition(current, expected):
    factory = _init_factory(current)
    assert factory.build_past_condition(current) == expected

@pytest.mark.parametrize(
    "current,expected",
    (   
        (
            datetime(2021, 12, 31, 23),
            GreaterThanOrEqualCondition(datetime(2021, 12, 31, 23))
        ),
        (
            datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            GreaterThanOrEqualCondition(datetime(2022, 5, 12, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")))
        ),
        (
            datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")), 
            GreaterThanOrEqualCondition(datetime(2022, 1, 31, 23, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")))
        ),
    )
)
def test_build_future_condition(current, expected):
    factory = _init_factory(current)
    assert factory.build_future_condition(current) == expected


@pytest.mark.parametrize(
    "current,operator,operator_to_mock,period,offset",
    (   
        (
            datetime(2025, 8, 10, tzinfo=zoneinfo.ZoneInfo("Africa/Malabo")),
            'today',
            'build_today_condition',
            1,
            0
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'yesterday',
            'build_yesterday_condition',
            2,
            1
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_x_days',
            'build_previous_x_days_condition',
            3,
            2
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_x_days_to_date',
            'build_previous_x_days_to_date_condition',
            4,
            3
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_week',
            'build_previous_week_condition',
            5,
            4
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_week_to_date',
            'build_previous_week_to_date_condition',
            6,
            5
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_month',
            'build_previous_month_condition',
            7,
            6
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_month_to_date',
            'build_previous_month_to_date_condition',
            8,
            7
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_quarter',
            'build_previous_quarter_condition',
            9,
            8
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_quarter_to_date',
            'build_previous_quarter_to_date_condition',
            10,
            9
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_year',
            'build_previous_year_condition',
            11,
            10
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'previous_year_to_date',
            'build_previous_year_to_date_condition',
            12,
            11
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'before_x_hours_ago',
            'build_before_x_hours_ago_condition',
            13,
            12
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'after_x_hours_ago',
            'build_after_x_hours_ago_condition',
            14,
            13
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'past',
            'build_past_condition',
            15,
            14
        ),
        (
            datetime(2022, 8, 10, tzinfo=zoneinfo.ZoneInfo("Europe/Paris")),
            'future',
            'build_future_condition',
            16,
            15
        ),
    )
)
def test_build(current, operator, operator_to_mock, period, offset):
    factory = _init_factory(current)
    with mock.patch.object(factory, operator_to_mock) as mocked:
        factory.build(operator, period, offset)
        mocked.assert_called_with(current_datetime=current, offset=offset, period=period)


import copy
from unittest import mock

from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
class ResourceCountViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json', 'session.json', 'question.json', 'choice.json']

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless_here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get(self, *args, **kwargs):
        url = reverse('django_forest:resources:count', kwargs={'resource': 'tests_question'})
        response = self.client.get(url)
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {'count': 3})

    @mock.patch('jose.jwt.decode', return_value={'id': 1})
    def test_get_invalid_token(self, *args, **kwargs):
        url = reverse('django_forest:resources:count', kwargs={'resource': 'tests_question'})
        response = self.client.get(url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'Missing required rendering_id'}]})

    def test_get_no_model(self, *args, **kwargs):
        url = reverse('django_forest:resources:count', kwargs={'resource': 'Foo'})
        response = self.client.get(url)
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource Foo'}]})



from django.test import TestCase, override_settings

from django_forest.utils.forest_setting import get_forest_setting

from test.support import EnvironmentVarGuard


class UtilsCorsTests(TestCase):

    def setUp(self):
        self.env = EnvironmentVarGuard()
        self.env.set('BAR', 'bar')

    @override_settings(FOREST={'FOO': 'foo'})
    def test_get_forest_setting(self):
        foo = get_forest_setting('FOO')
        self.assertEqual(foo, 'foo')

    def test_get_forest_setting_none(self):
        foo = get_forest_setting('FOO')
        self.assertEqual(foo, None)

    def test_get_forest_setting_default(self):
        foo = get_forest_setting('FOO', True)
        self.assertEqual(foo, True)

    @override_settings(FOREST={'FOO': 'True'})
    def test_get_forest_setting_bool(self):
        foo = get_forest_setting('FOO', False)
        self.assertEqual(foo, True)

    @override_settings(FOREST={'FOO': 'foo'})
    def test_get_forest_setting_bad_bool(self):
        foo = get_forest_setting('FOO', False)
        self.assertEqual(foo, False)

    def test_get_forest_setting_env(self):
        bar = get_forest_setting('BAR')
        self.assertEqual(bar, 'bar')

    @override_settings(FOREST={'BAR': 'new_bar'})
    def test_get_forest_setting_env_setting(self):
        bar = get_forest_setting('BAR')
        self.assertEqual(bar, 'new_bar')



import requests

from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester


def retrieve():
    response = ForestApiRequester.get('/oidc/.well-known/openid-configuration')
    if response.status_code == requests.codes.ok:
        return response.json()
    else:
        raise Exception(MESSAGES['SERVER_TRANSACTION']['OIDC_CONFIGURATION_RETRIEVAL_FAILED'])



import copy
import json
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.models import Question, Restaurant
from django_forest.tests.resources.views.list.test_list_scope import mocked_scope
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
@mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})    
class ResourceListViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json',
                'session.json',
                'question.json', 'choice.json',
                'place.json', 'restaurant.json',
                'student.json',
                'serial.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self, *args, **kwargs):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = f"{reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})}?timezone=Europe%2FParis"
        self.reverse_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_choice'})
        self.no_data_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_waiter'})
        self.enum_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_student'})
        self.uuid_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_serial'})
        self.one_to_one_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_restaurant'})
        self.bad_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_foo'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self, *args, **kwargs):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    def test_get(self, *args, **kwargs):
        response = self.client.get(self.url, {'page[number]': '1', 'page[size]': '15'})
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/1/relationships/choice_set'
                            }
                        },
                        'topic': {'data': None, 'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                    },
                    'attributes': {
                        'question_text': 'what is your favorite color?',
                        'pub_date': '2021-06-02T13:52:53.528000+00:00'},
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                }, {
                    'type': 'tests_question',
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/2/relationships/choice_set'
                            }
                        },
                        'topic': {'data': None, 'links': {'related': '/forest/tests_question/2/relationships/topic'}}
                    },
                    'attributes': {
                        'question_text': 'do you like chocolate?',
                        'pub_date': '2021-06-02T15:52:53.528000+00:00'
                    },
                    'id': 2, 'links': {
                        'self': '/forest/tests_question/2'
                    }
                },
                {
                    'type': 'tests_question',
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/3/relationships/choice_set'
                            }
                        },
                        'topic': {'data': None, 'links': {'related': '/forest/tests_question/3/relationships/topic'}}
                    },
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    }
                }
            ]
        })

    
    def test_get_no_data(self, *args, **kwargs):
        response = self.client.get(self.no_data_url)
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {'data': []})

    def test_get_no_model(self, *args, **kwargs):
        response = self.client.get(self.bad_url, {'page[number]': '1', 'page[size]': '15'})
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})

    def test_post(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'question_text': 'What is your favorite color',
                    'pub_date': '2021-06-21T09:46:39.000Z'
                }
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.post(self.url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/4/relationships/choice_set'
                        }
                    },
                    'topic': {'links': {'related': '/forest/tests_question/4/relationships/topic'}}
                },
                'attributes': {
                    'pub_date': '2021-06-21T09:46:39+00:00',
                    'question_text': 'What is your favorite color'
                },
                'id': 4,
                'links': {
                    'self': '/forest/tests_question/4'
                }
            },
            'links': {
                'self': '/forest/tests_question/4'
            }
        })
        self.assertEqual(Question.objects.count(), 4)

    def test_post_no_model(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'question_text': 'What is your favorite color',
                    'pub_date': '2021-06-21T09:46:39.000Z'
                }
            }
        }
        response = self.client.post(self.bad_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})

    def test_post_related_data(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'serves_hot_dogs': False,
                    'serves_pizzas': False
                },
                'relationships': {
                    'place': {
                        'data': {
                            'type': 'tests_places',
                            'id': 2,
                        }
                    },
                }
            }
        }
        self.assertEqual(Restaurant.objects.count(), 1)
        response = self.client.post(self.one_to_one_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_restaurant',
                'relationships': {
                    'waiter_set': {
                        'links': {
                            'related': '/forest/tests_restaurant/2/relationships/waiter_set'
                        }
                    },
                    'place': {
                        'links': {
                            'related': '/forest/tests_restaurant/2/relationships/place'
                        }
                    }
                },
                'attributes': {
                    'serves_hot_dogs': False,
                    'serves_pizza': False
                },
                'id': 2,
                'links': {
                    'self': '/forest/tests_restaurant/2'
                }
            },
            'links': {
                'self': '/forest/tests_restaurant/2'
            }
        })
        self.assertEqual(Restaurant.objects.count(), 2)

    def test_post_related_data_do_not_exist(self, *args, **kwargs):
        body = {
            'data': {
                'attributes': {
                    'serves_hot_dogs': False,
                    'serves_pizzas': False
                },
                'relationships': {
                    'place': {
                        'data': {
                            'type': 'tests_places',
                            'id': 3,
                        }
                    },
                }
            }
        }
        self.assertEqual(Restaurant.objects.count(), 1)
        response = self.client.post(self.one_to_one_url,
                                    json.dumps(body),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Instance tests_place with pk 3 does not exists'
                }
            ]
        })
        self.assertEqual(Restaurant.objects.count(), 1)

    def test_post_error(self, *args, **kwargs):
        url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        data = {
            'data': {
                'attributes': {
                    'question_text': '''aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                    aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                    aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                    aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                    a''',
                }
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.post(url,
                                    json.dumps(data),
                                    content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'value too long for type character varying(200)\n'
                }
            ]
        })
        self.assertEqual(Question.objects.count(), 3)

    def test_delete(self, *args, **kwargs):
        data = {
            'data': {
                'attributes': {
                    'ids': ['1'],
                }
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.delete(self.url,
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Question.objects.count(), 2)

    def test_delete_scope(self, *args, **kwargs):
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': 'useless-here'
            }
        }
        data = {
            'data': {
                'attributes': {
                    'ids': ['2'],
                }
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.delete(self.url,
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Question.objects.count(), 3)

    def test_delete_all_records(self, *args, **kwargs):
        data = {
            'data': {
                'attributes': {
                    'ids': [
                        '1'
                    ],
                    'collection_name': 'tests_question',
                    'parent_collection_name': None,
                    'parent_collection_id': None,
                    'parent_association_name': None,
                    'all_records': True,
                    'all_records_subset_query': {
                        'fields[tests_question]': 'id,question_text,pub_date',
                        'fields[subject]': 'name',
                        'fields[subject2]': 'name',
                        'fields[topic]': 'name',
                        'page[number]': 1,
                        'page[size]': 15,
                        'sort': '-id',
                        'filters': '{\"field\":\"question_text\",\"operator\":\"equal\",\"value\":\"what is your favorite color?\"}',
                        "searchExtended": 0
                    },
                    'all_records_ids_excluded': [],
                    'smart_action_id': None
                },
                'type': 'action-requests'
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.delete(self.url,
                                      json.dumps(data),
                                      content_type='application/json')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Question.objects.count(), 2)

    def test_delete_no_model(self, *args, **kwargs):
        data = {
            'data': {
                'attributes': {
                    'ids': ['1'],
                }
            }
        }
        self.assertEqual(Question.objects.count(), 3)
        response = self.client.delete(self.bad_url,
                                      json.dumps(data),
                                      content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'no model found for resource tests_foo'}]})
        self.assertEqual(Question.objects.count(), 3)



"""
Django settings for foo project.
Generated by 'django-admin startproject' using Django 3.2.2.
For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/
For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""
import base64
import os
from pathlib import Path
import uuid

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'django-insecure-o2h_(oc$6trs!=r&-vx2+8pst0_^ehq&m7ua$#i&_a0jh_-ji%'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django_forest.tests.apps.TestsConfig',
    'django.contrib.auth',
    'django.contrib.contenttypes',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_forest.tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR / 'templates'],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'foo.wsgi.application'


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': os.getenv('POSTGRES_DB', 'django_forest'),
        'USER': os.getenv('POSTGRES_USER', 'forest'),
        'PASSWORD': os.getenv('POSTGRES_PASSWORD', 'secret'),
        'HOST': os.getenv('POSTGRES_HOST', 'localhost'),
        'PORT': os.getenv('POSTGRES_PORT', '5447'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = '/static/'

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

FOREST = {
    'FOREST_URL': 'https://api.test.forestadmin.com',
    'APPLICATION_URL': 'http://localhost:8000',
    'FOREST_ENV_SECRET': 'foo',
    'FOREST_AUTH_SECRET': base64.urlsafe_b64encode(uuid.uuid4().bytes),
    # 'INCLUDED_MODELS': [],
    # 'EXCLUDED_MODELS': ['Permission', 'Group', 'User', 'ContentType'],
    # 'CONFIG_DIR': 'forest',
    # 'FOREST_DISABLE_AUTO_SCHEMA_APPLY': True
}

APPEND_SLASH = False



from django.http import JsonResponse, HttpResponse

from django_forest.resources.utils.format import FormatFieldMixin
from django_forest.resources.utils.json_api_serializer import JsonApiSerializerMixin
from django_forest.resources.utils.resource import ResourceView
from django_forest.resources.utils.smart_field import SmartFieldMixin
from django_forest.utils.schema.json_api_schema import JsonApiSchema


class ListView(FormatFieldMixin, SmartFieldMixin, JsonApiSerializerMixin, ResourceView):
    def get(self, request):
        # default
        queryset = self.Model.objects.all()

        params = request.GET.dict()

        try:
            # enhance queryset
            queryset = self.enhance_queryset(queryset, self.Model, params, request)

            # handle smart fields
            self.handle_smart_fields(queryset, self.Model._meta.db_table, many=True)

            # json api serializer
            data = self.serialize(queryset, self.Model, params)

            # search decorator
            data = self.decorators(data, self.Model, params)
        except Exception as e:
            return self.error_response(e)
        else:
            return JsonResponse(data, safe=False)

    def post(self, request):
        body = self.get_body(request.body)

        try:
            attributes = self.populate_attribute(body)
            instance = self.Model.objects.create(**attributes)
        except Exception as e:
            return self.error_response(e)
        else:
            # json api serializer
            Schema = JsonApiSchema._registry[f'{self.Model._meta.db_table}Schema']
            data = Schema().dump(instance)
            return JsonResponse(data, safe=False)

    def delete(self, request):
        queryset = self.Model.objects.all()
        scope_filters = self.get_scope(request, self.Model)
        if scope_filters is not None:
            queryset = queryset.filter(scope_filters)

        ids = self.get_ids_from_request(request, self.Model)
        # Notice: this does not run pre/post_delete signals
        queryset.filter(pk__in=ids).delete()
        return HttpResponse(status=204)



from django_forest.utils.schema import Schema


def date_difference_in_seconds(date1, date2):
    return (date1 - date2).total_seconds()


def is_stat_permission_allowed(array_permission_infos, pool_permission):
    for info in array_permission_infos:
        if info not in pool_permission:
            return False
    else:
        return True


def is_stat_allowed(pool_permissions, array_permission_infos):
    # NOTICE: Is there any pool_permissions containing the array_permission_infos
    for pool_permission in pool_permissions:
        pool_permission = [x for x in pool_permission.values() if x is not None]
        if is_stat_permission_allowed(array_permission_infos, pool_permission):
            return True
    else:
        return False


def is_user_allowed(user_id, permission_value):
    if permission_value is None:
        return False
    elif permission_value in (True, False):
        return permission_value
    else:
        return user_id in permission_value


def find_action_from_endpoint(obj):
    endpoint = obj.smart_action_request_info['endpoint']
    http_method = obj.smart_action_request_info['http_method']
    collection = Schema.get_collection(obj.collection_name)
    return next((x for x in collection['actions']
                 if x['endpoint'] == endpoint and x['http_method'] == http_method),
                None)


def is_smart_action_allowed(obj, smart_actions_permissions):
    schema_smart_action = find_action_from_endpoint(obj)

    try:
        permission = smart_actions_permissions[schema_smart_action['name']]
    except Exception:
        return False
    else:
        return is_user_allowed(obj.user_id, permission['triggerEnabled'])



from django_forest.tests.models import Choice, Topic
from django_forest.utils.collection import Collection


class ChoiceForest(Collection):
    def load(self):
        self.fields = [
            {
                'field': 'topic',
                'reference': 'tests_topic.id',
                'type': 'String',
                'get': self.get_topic,
            }
        ]

    def get_topic(self, obj):
        queryset = Topic.objects.filter(question__in=(obj.question_id,))
        if len(queryset):
            return queryset[0]


Collection.register(ChoiceForest, Choice)



# Generated by Django 3.2.4 on 2021-08-06 22:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0005_alter_serial_uuid'),
    ]

    operations = [
        migrations.CreateModel(
            name='Topic',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
            ],
        ),
        migrations.AddField(
            model_name='question',
            name='topic',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='tests.topic'),
        ),
    ]



from .filters import ConditionsMixin
from django_forest.utils import get_token
from django_forest.utils.scope import ScopeManager
from django_forest.utils.date import get_timezone


class ScopeMixin(ConditionsMixin):
    def get_scope(self, request, Model):
        token = get_token(request)
        filters = ScopeManager.get_scope_for_user(token, Model._meta.db_table)
        if filters is not None:
            tz = get_timezone(request.GET['timezone'])
            return self.handle_aggregator(filters, Model, tz)



import copy
import json
import sys

import pytest
import pytz

from datetime import datetime
from unittest import mock

from django.conf import settings
from django.test import TestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.middleware import mocked_permissions, mocked_config, \
    mocked_requests_permission
from django_forest.utils.middlewares import set_middlewares
from django_forest.utils.permissions import Permission
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager

mocked_permissions_scope = copy.deepcopy(mocked_permissions)
mocked_permissions_scope['data']['renderings']['1'] = {
    'tests_question': {
        'scope': {
            'filter': {
                'aggregator': 'and',
                'conditions': [
                    {
                        'field': 'question_text',
                        'operator': 'contains',
                        'value': 'what'
                    }
                ]},
            'dynamicScopesValues': {}
        }
    }
}

mocked_permissions_scope_dynamic = copy.deepcopy(mocked_permissions_scope)
mocked_permissions_scope_dynamic['data']['renderings']['1']['tests_question']['scope']['filter']['conditions'][0][
    'value'] = '$currentUser.firstName'
mocked_permissions_scope_dynamic['data']['renderings']['1']['tests_question']['scope']['dynamicScopesValues'] = {
    'users': {
        '1': {
            '$currentUser.firstName': 'Guillaume'
        }
    }
}

mocked_permissions_none = copy.deepcopy(mocked_permissions)
mocked_permissions_none['data']['collections']['tests_question']['collection']['browseEnabled'] = None

mocked_permissions_user = copy.deepcopy(mocked_permissions)
mocked_permissions_user['data']['collections']['tests_question']['collection']['browseEnabled'] = ['1']

mocked_permissions_no_collection = copy.deepcopy(mocked_permissions)
del mocked_permissions_no_collection['data']['collections']['tests_question']

mocked_permissions_list_forbidden = copy.deepcopy(mocked_permissions)
mocked_permissions_list_forbidden['data']['collections']['tests_question']['collection']['browseEnabled'] = False

mocked_permissions_action = copy.deepcopy(mocked_permissions)
mocked_permissions_action['data']['collections']['tests_question']['actions'] = {
    'Send invoice': {
        'triggerEnabled': True
    }
}

mocked_permissions_stats = copy.deepcopy(mocked_permissions)
mocked_permissions_stats['stats'] = {
    'queries': ['SELECT COUNT(*) AS value, 5 as objective FROM tests_question',
                'SELECT SUM(tests_question.id) AS value FROM tests_question'],
    'leaderboards': [{'type': 'Leaderboard', 'limit': 5, 'aggregator': 'Count', 'labelFieldName': 'question_text',
                      'aggregateFieldName': None, 'sourceCollectionId': 'tests_question',
                      'relationshipFieldName': 'choice_set'}],
    'lines': [{'type': 'Line', 'filter': '{"field":"id","operator":"equal","value":0}', 'timeRange': 'Day',
               'aggregator': 'Count', 'groupByFieldName': 'pub_date', 'aggregateFieldName': None,
               'sourceCollectionId': 'tests_question'}],
    'objectives': [],
    'percentages': [{'type': 'Percentage', 'numeratorChartId': 'ffd97b70-e3b9-11eb-aaf7-0ffaf80df470',
                     'denominatorChartId': '49dd5600-e587-11eb-aaf6-e990d6816f15'}],
    'pies': [
        {'type': 'Pie', 'filter': None, 'aggregator': 'Sum', 'groupByFieldName': 'pub_date', 'aggregateFieldName': 'id',
         'sourceCollectionId': 'Question'}],
    'values': [{'type': 'Value', 'filter': '{"field":"pub_date","operator":"previous_month","value":null}',
                'aggregator': 'Sum', 'aggregateFieldName': 'id', 'sourceCollectionId': 'tests_question'}]
}

mocked_permissions_missing_stats = copy.deepcopy(mocked_permissions)
mocked_permissions_missing_stats['stats'] = {}

mocked_config_bad_request = copy.deepcopy(mocked_config)
mocked_config_bad_request['permissions'] = {
    'data': {},
    'status': 400
}
mocked_config_no_collection = copy.deepcopy(mocked_config)
mocked_config_no_collection['permissions']['data'] = mocked_permissions_no_collection
mocked_config_none = copy.deepcopy(mocked_config)
mocked_config_none['permissions']['data'] = mocked_permissions_none
mocked_config_user = copy.deepcopy(mocked_config)
mocked_config_user['permissions']['data'] = mocked_permissions_user
mocked_config_list_forbidden = copy.deepcopy(mocked_config)
mocked_config_list_forbidden['permissions']['data'] = mocked_permissions_list_forbidden
mocked_config_action = copy.deepcopy(mocked_config)
mocked_config_action['permissions']['data'] = mocked_permissions_action
mocked_config_stats = copy.deepcopy(mocked_config)
mocked_config_stats['permissions']['data'] = mocked_permissions_stats
mocked_config_missing_stats = copy.deepcopy(mocked_config)
mocked_config_missing_stats['permissions']['data'] = mocked_permissions_missing_stats


class MiddlewarePermissionsNoTokenTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')
        settings.MIDDLEWARE.remove('django_forest.middleware.IpWhitelistMiddleware')

    def test_list(self):
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 403)


class MiddlewarePermissionsTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')
        settings.MIDDLEWARE.remove('django_forest.middleware.IpWhitelistMiddleware')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config))
    def test_list(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15',
        })
        self.assertEqual(response.status_code, 200)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_bad_request))
    def test_list_error_server(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15',
        })
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_no_collection))
    def test_list_no_collection(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_none))
    def test_list_none(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_user))
    def test_list_user(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 200)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_list_forbidden))
    def test_list_forbidden(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 403)


class MiddlewarePermissionsCookieTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.client = self.client_class(
            HTTP_COOKIE='forest_session_token=eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')
        settings.MIDDLEWARE.remove('django_forest.middleware.IpWhitelistMiddleware')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config))
    def test_list(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 200)


class MiddlewarePermissionsCachedTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        Permission.roles_acl_activated = True
        Permission.permissions_cached = {
            'data': {
                'collections': {
                    'tests_question': {
                        'collection': {
                            'browseEnabled': True,
                            'readEnabled': True,
                            'addEnabled': True,
                            'editEnabled': True,
                            'deleteEnabled': True,
                            'exportEnabled': True
                        },
                        'actions': {}
                    }
                },
                'renderings': {
                    '1': {
                        'stats': {
                            'queries': [],
                            'leaderboards': [],
                            'lines': [],
                            'objectives': [],
                            'percentages': [],
                            'pies': [],
                            'values': []
                        },
                        'last_fetch': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
                    }
                }
            },
            'stats': {
                'queries': [],
                'leaderboards': [],
                'lines': [],
                'objectives': [],
                'percentages': [],
                'pies': [],
                'values': []
            },
            'meta': {
                'rolesACLActivated': True
            },
            'last_fetch': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        }
        Permission.renderings_cached = {
            '1': {
                'stats': {
                    'queries': [],
                    'leaderboards': [],
                    'lines': [],
                    'objectives': [],
                    'percentages': [],
                    'pies': [],
                    'values': []
                },
                'last_fetch': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')
        settings.MIDDLEWARE.remove('django_forest.middleware.IpWhitelistMiddleware')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config))
    @mock.patch('django_forest.utils.permissions.Permission.fetch_permissions')
    def test_list_once_again(self, mocked_fetch_permissions, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 200)
        self.assertFalse(mocked_fetch_permissions.called)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config))
    @mock.patch('django_forest.utils.permissions.Permission.fetch_permissions')
    def test_list_no_last_fetch_renderings_cached(self, mocked_fetch_permissions, mocked_requests, mocked_datetime,
                                                  mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        del Permission.renderings_cached['1']['last_fetch']
        response = self.client.get(self.url, {
            'page[number]': '1',
            'page[size]': '15'
        })
        self.assertEqual(response.status_code, 200)
        self.assertFalse(mocked_fetch_permissions.called)


# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class MiddlewarePermissionsActionsTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('actions:send-invoice')
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        self.body = {
            'data': {
                'attributes': {
                    'collection_name': 'tests_question',
                    'values': {},
                    'ids': [
                        '1'
                    ],
                    'parent_collection_name': None,
                    'parent_collection_id': None,
                    'parent_association_name': None,
                    'all_records': False,
                    'all_records_subset_query': {
                        'fields[pollstests_question]': 'id,pubDate,questionText',
                        'fields[subject]': 'name',
                        'fields[subject2]': 'name',
                        'fields[topic]': 'name',
                        'page[number]': 1,
                        'page[size]': 15,
                        'sort': '-id',
                        'searchExtended': 0,
                        'timezone': 'Europe/Paris'
                    },
                    'all_records_ids_excluded': [
                        '2',
                        '3'
                    ],
                    'smart_action_id': 'tests_question-Send@@@Invoice@@@'
                },
                'type': 'custom-action-requests'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')
        settings.MIDDLEWARE.remove('django_forest.middleware.IpWhitelistMiddleware')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_action))
    def test_actions(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(self.url, json.dumps(self.body), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data, {'success': 'now live'})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_action))
    def test_actions_not_exist(self, mocked_requests, mocked_datetime, mocked_decode):
        url = reverse('actions:not-exists')
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(url, json.dumps(self.body), content_type='application/json')
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_action))
    def test_actions_no_resource(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        body = copy.deepcopy(self.body)
        body['data']['attributes']['collection_name'] = 'Foo'
        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        self.assertEqual(response.status_code, 400)


class MiddlewarePermissionsStatsTests(TestCase):

    def setUp(self):
        set_middlewares()
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.live_queries_url = f"{reverse('django_forest:stats:liveQueries')}?timezone=Europe%2FParis"
        self.stats_with_parameters_url = f"{reverse('django_forest:stats:statsWithParameters', kwargs={'resource': 'tests_question'})}?timezone=Europe%2FParis"
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        self.live_queries_body = {
            'type': 'Objective',
            'timezone': 'Europe/Paris',
            'query': 'SELECT COUNT(*) AS value, 5 as objective FROM tests_question'
        }
        self.stats_with_parameters_body = {
            'type': 'Leaderboard',
            'collection': 'tests_question',
            'timezone': 'Europe/Paris',
            'label_field': 'question_text',
            'relationship_field': 'choice_set',
            'limit': 5,
            'aggregate': 'Count'
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Permission.permissions_cached = {}
        Permission.renderings_cached = {}
        ScopeManager.cache = {}
        settings.MIDDLEWARE.remove('django_forest.middleware.PermissionMiddleware')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_stats))
    def test_live_queries(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(self.live_queries_url, json.dumps(self.live_queries_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_stats))
    def test_live_queries_forbidden(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        live_queries_body = {
            'type': 'Objective',
            'timezone': 'Europe/Paris',
            'query': 'SELECT COUNT(*) AS value, 0 as objective FROM tests_question'
        }

        response = self.client.post(self.live_queries_url, json.dumps(live_queries_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_missing_stats))
    def test_live_queries_missing_stats(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(self.live_queries_url, json.dumps(self.live_queries_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_stats))
    def test_stats_with_parameters(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(self.stats_with_parameters_url, json.dumps(self.stats_with_parameters_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_stats))
    def test_stats_with_parameters_forbidden(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        stats_with_parameters_body = {
            'type': 'Value',
            'aggregator': 'Count',
            'sourceCollectionId': 'tests_question',
            'timezone': 'Europe/Paris'
        }
        response = self.client.post(self.stats_with_parameters_url, json.dumps(stats_with_parameters_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('requests.get', side_effect=mocked_requests_permission(mocked_config_missing_stats))
    def test_stats_with_parameters_missing_stats(self, mocked_requests, mocked_datetime, mocked_decode):
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        response = self.client.post(self.stats_with_parameters_url, json.dumps(self.stats_with_parameters_body),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)



import logging

from django.http import JsonResponse

from django_forest.utils.scope import ScopeManager

# Get an instance of a logger
from django_forest.utils.views.base import BaseView

logger = logging.getLogger(__name__)


class ScopeCacheInvalidationView(BaseView):
    def post(self, request, *args, **kwargs):

        if not self.is_authenticated(request):
            return JsonResponse({'errors': [{'detail': 'Please authenticate'}]}, status=403)

        body = self.get_body(request.body)
        if 'renderingId' not in body:
            msg = 'missing renderingId'
            logger.error(msg)
            return self.error_response(msg)

        ScopeManager.invalidate_scope_cache(str(body['renderingId']))
        return JsonResponse({}, status=200)
